# core_ops.tcl -- all the backstage stuff.

set shape {A B C D E F}
set fill {a b c d e f}
set X {1 2 3 4 5 6 7 8 9 10 11 12}
set Y {1 2 3 4 5 6 7 8}

proc get_board {x y} { global board; return $board($x\ $y) }
proc set_board {x y v} { global board neigh
        set board($x\ $y) $v 
        foreach n $neigh($x\ $y) { eval set_nsew $n }
}
proc set_nsew {x y} { global board neigh nsew
        set list {}
        foreach n $neigh($x\ $y) { 
                set v $board($n); if {$v != {}} {lappend list $v}
        }; set nsew($x\ $y) $list
}
proc get_nsew {x y} { global nsew; return $nsew($x\ $y) }

proc init_board {} { 
	uplevel "#0" { 
                catch {unset board}; catch {unset neigh}; catch {unset nsew}
                set s_un {}; set s_re {}
        }
        global board neigh nsew X Y
        foreach x $X {
                foreach y $Y { set xy $x\ $y
			set board($xy) {}; set nsew($xy) {}
                        incr x -1; if {$x > 0} { lappend neigh($x\ $y) $xy }
                        incr x 2;  if {$x <13} { lappend neigh($x\ $y) $xy }
                        incr x -1
                        incr y -1; if {$y > 0} { lappend neigh($x\ $y) $xy }
                        incr y 2;  if {$y < 9} { lappend neigh($x\ $y) $xy }
                        incr y -1
                }
        }
}
proc init_pouch {} { global prepouch pouch shape fill 
        save_random
	set prepouch {}; set pouch {}; set ts $shape; set tf $fill 
	foreach s $shape { foreach f $fill { set pool([list $s $f]) 2 } }
	foreach i {6 5 4 3 2 1} {
		set rs [random $i 0]; set rf [random $i 0]
		set s [lindex $ts $rs]; set f [lindex $tf $rf]
		set ts [lreplace $ts $rs $rs]; set tf [lreplace $tf $rf $rf]
		lappend prepouch [list $s $f]; incr pool([list $s $f]) -1
	}
	set i 0; foreach v [array names pool] {
		while {$pool($v)} { incr i; incr pool($v) -1
			set pouch [linsert $pouch [random $i] $v]
		} 
	}; if {[llength $pouch] != 66} { error "Error creating pouch." }
        finish_random
}

proc _score {{p {}} {n {}}} { upvar "#0" player($p,score) s player($p,four) f
	switch -- $n {
	{} { return [list $s $f] }
	1 { incr s 1 }
	2 { incr s 2 }
	3 { incr s 4 }
	4 { incr f 1; incr s [expr {1 << (2+$f)}] }
	default {error "_score: invalid arg"}
	}
}

proc init_player {p} { upvar "#0" player($p,score) s player($p,four) f
	set s 0; set f 0
}

proc _place {x y v {p {}}} { global s_un s_re
	set res [does_match $x $y $v]
	if {$res == 0} { invalid; return 0} 
	if {$x==1 || $x==12 || $y==1 || $y==8} {set res 0}

        set args [list $x $y $v $p]; set oldscore [_score $p]
	if {$args == [lindex $s_re end]} {
	        set s_re [lreplace $s_re end end] 
	} else { set s_re {} }
	set_board $x $y $v; if {$res > 0} { _score $p $res }
	lappend s_un [concat $args $oldscore]
	return 1
}

proc _undo {} { global s_un s_re
	set t [lindex $s_un end]
	set s_un [lreplace $s_un end end] 
	foreach {x y v p os of} $t {}
	lappend s_re [lrange $t 0 3]
	if {[get_board $x $y] != $v} { error "Stack-error" }
	upvar "#0" player($p,score) s player($p,four) f
	set s $os; set f $of; set_board $x $y {}
}

proc _redo {} { global s_re
	eval _place [lindex $s_re end] 
}

proc get_undo_info {} { global s_un
	return [lrange [lindex $s_un end] 0 3]
}
proc get_redo_info {} { global s_re
	return [lindex $s_re end]
}

proc get_count {nb v} { 
	foreach {c1 c2} {0 0}  {v1 v2} $v {}
	foreach n $nb {
		foreach {n1 n2} $n  any 0 {}
		if {$n1 == $v1} { incr c1; set any 1}
		if {$n2 == $v2} { incr c2; set any 1}
		if {$any != 1} { return {0 0} }
	}
	return [list $c1 $c2]
}
proc does_match {x y v} {
	if {[get_board $x $y] != {}} {return 0}
	set nsew [get_nsew $x $y]; if {$nsew == {}} {return 0}
	set count [get_count $nsew $v]
	foreach {c1 c2} $count {}
	switch -- [llength $nsew] {
        0 {return 0}
	1 { if {$c1 > 0 || $c2 > 0} {return 1} {return 0}}
	2 { if {$c1 > 0 && $c2 > 0} {return 2} {return 0}}
	3 { if {$c1 > 0 && $c2 > 0} {return 3} {return 0}}
	4 { if {$c1 > 1 && $c2 > 1} {return 4} {return 0}}
	}
}

proc search_pos {v} { global X Y
	set matchlist {}
	foreach x $X {
		foreach y $Y {
			if {[does_match $x $y $v]} {
				lappend matchlist [list $x $y]
			}
		}
	}
	return $matchlist
}

proc invalid {} { puts "Invalid" }

proc start_game {} { global prepouch
	init_board 
	set_board 1 1 [lindex $prepouch 0]
	set_board 12 1 [lindex $prepouch 1] 
	set_board 12 8 [lindex $prepouch 2] 
	set_board 1 8 [lindex $prepouch 3] 
	set_board 6 4 [lindex $prepouch 4] 
	set_board 7 5 [lindex $prepouch 5] 
}
proc new_game {} {
	init_pouch
	start_game
}
proc _pouch {{i {}}} { global pouch prepouch
	if {$i == "pre"} { return $prepouch } 
	if {$i == {}} { return $pouch } else { return [lindex $pouch $i] }
}
proc _current {} { global s_un; return [llength $s_un] }

# initialize
# init_player {}
new_game

