#!/bin/sh
# \
exec wish "$0" {1+$@}

# gui.tcl -- all the GUI-stuff:

source "random.tcl"
source "core_ops.tcl"
source "dialog.tcl"
source "images.tcl"
source "rules.tcl"

proc xy_to_wn {x y {subw {.l}}} { return .board.$x,$y$subw }
proc wn_to_xy {w} { return [split [winfo name [winfo parent $w]] ,] }

proc set_image {x y} { global img 
    set image $img([get_board $x $y])
    [xy_to_wn $x $y] conf -image $image
}

proc update_score {{p {}}} {
    foreach {s f} [_score $p] {}
    .board.score conf -text "$p:   Score: $s   4Ways: $f"
}


proc invalid {} { global option
    if {$option(sound)} { bell }
}

proc win_place {w} { global cur_stone cur_player
    board_focus $w; show_img $w 
    if {$cur_stone == {}} {return}
    set xy [wn_to_xy $w]
    if [eval _place $xy [list $cur_stone $cur_player]] {
        eval set_image $xy; update_score $cur_player
        set cur_stone {}; show_stone
        # next_player $cur_player
    }
}

proc win_undo {} { 
    foreach {x y v p} [get_undo_info] {
        _undo; set_image $x $y; show_stone 
        update_score $p; board_focus $x $y
        global cur_player; set cur_player $p
    }
}

proc win_redo {} { 
    foreach {x y v p} [get_redo_info] {
        _redo; set_image $x $y; show_stone 
        update_score $p; # next_player $p
        foreach {x y v p} [get_redo_info] {}
        board_focus $x $y
    }
}

proc win_new_game {} { global X Y option
    new_game; foreach x $X { foreach y $Y { set_image $x $y } }
    do_pouch; show_stone
}

proc win_undo_all {} {
    while {[get_undo_info] != {}} { win_undo; update }
}

proc win_redo_all {} {
    while {[get_redo_info] != {}} { win_redo; update }
}

proc show_stone {} { global img cur_stone 
    set cur_stone [_pouch [_current]]
    set newimg $img($cur_stone)
    .board.next conf -image $newimg
    .board.bubble conf -image $newimg
    if {[get_undo_info] == {}} { .cmds.undo conf -state disabled
    } else { .cmds.undo conf -state normal }
    if {[get_redo_info] == {}} { .cmds.redo conf -state disabled
    } else { .cmds.redo conf -state normal }
    update_pouch
    move_possible $cur_stone
}    

proc move_possible {v} {
    set c [_current]
    set l [expr 66 - $c]

    if {$v == {}} { 
        .board.infos conf -text "done\nall\n\ntodo\nnone\n\n\n"
        finished
    } else { 
        global cmd_cnt; set old [info cmdcount]
        set poslist [search_pos $v] 
        set cmd_cnt [expr [info cmdcount] - $old]

        set m [llength $poslist]
        .board.infos conf -text "done\n$c\n\ntodo\n$l\n\nways\n$m"
        if {$poslist == {}} { no_more } 
    }
}

proc no_more {} { global cur_stone img 
    set res [ img_dialog .mess "no matches" "There are\nno matches"\
                $img($cur_stone) 0 \
                "Ok" "Undo Last" "Start Over" "New Game"]
    switch -- $res {
    0 {}
    1 { win_undo }
    2 { win_undo_all }
    3 { win_new_game }
    default {}
    }
}
proc finished {} { global within_finished
    if {$within_finished} return
    set within_finished 1
    win_undo_all; after 500; win_redo_all
    set within_finished 0

    set message ""; append message \
        "Congratulations !\n" \
        "You've done it !\n" \
        "You're a Genius !\n" \
        "You get a zillion \n" \
        "   Bonus-points ...\n" \
        "... and be happy for \n" \
        "   the rest of your life."
    set res [ tk_dialog .mess "Finished" $message \
            warning 0 "Ok, I'm happy." "Start Over" "New game"]
    switch -- $res {
    0 {}
    1 { win_undo_all }
    2 { win_new_game }
    default {}
    }
}

proc board_focus {{x {}} {y {}}} { global board_focus
    if {$x != {} && $y != {}} {
        set board_focus [list $x $y]
        focus [xy_to_wn $x $y]
    } elseif {$x == "get"} { return $board_focus 
    } elseif {$x != {}} {
        eval board_focus [wn_to_xy $x]
    } else { focus [eval xy_to_wn $board_focus] }
}
        
proc board_key_handler {w k} { 
    foreach {x y} [wn_to_xy $w] {}
    if {$y == {}} { foreach {x y} [board_focus get] {} }
    
    set used 1
    switch -- $k {
    space  { win_place $w }
    Return { win_place $w }
    k - K - Up     { if {$y > 1} { incr y -1 } }
    j - J - Down   { if {$y < 8} { incr y  1 } }
    h - H - Left   { if {$x > 1} { incr x -1 } }
    l - L - Right  { if {$x < 12} { incr x  1 } }
    default { set used 0 }
    }
    board_focus $x $y
    if {$used} { return -code break }
}
proc key_handler {w k} {
    switch -- $k {
    h - F1 { display_help }
    g      { win_new_game }
    i      { display_random }
    u      { win_undo }
    U      { win_undo_all }
    r      { win_redo }
    R      { win_redo_all }
    q      { destroy [winfo toplevel $w] }
    Escape { hide_img }
    default {}
    }
}
proc show_img {{w {}}} { global offsx offsy img_shown option
    if {$w == {}} { set w $img_shown } else { set img_shown $w }
    if {$option(bubble) == 0 || $w == {}} return

    set w [winfo parent $w]
    set x [winfo x $w]; set y [winfo y $w]
    incr x $offsx; incr y $offsy
    place .board.bubble -x $x -y $y; raise .board.bubble
}
proc hide_img {{w {}}} {
    if {$w != {}} { uplevel "#0" { set img_shown {} } }
    lower .board.bubble; place forget .board.bubble
}
proc do_img {} { global option
    if {$option(bubble)} { show_img } { hide_img }
}
proc do_pouch {{o {}}} { global option img 
    if {$o != {}} { set option(pouch) $o }
    set cur [expr {[_current] + 6}]
    if {$option(pouch)} {
        if {[winfo exists .pouch]} {
            eval destroy [winfo children .pouch]
        } else { toplevel .pouch }
        bind .pouch <Destroy> {if {"%W" == ".pouch"} {set option(pouch) 0}}
        set i 0; set j 0; set n 0
        foreach v  [_pouch pre]  {
            label .pouch.$n -image $img($v) \
                            -relief flat -bd 1
            grid .pouch.$n -row $j -column $i
            incr i; incr n; if {$i==6} {set i 0; incr j}
        }
        foreach v  [_pouch]  {
            label .pouch.$n -image $img($v) \
                            -relief raised -bd 1
            grid .pouch.$n -row $j -column $i
            incr i; incr n; if {$i==6} {set i 0; incr j}
        }
        update_pouch
    } else {
        catch {destroy .pouch}
    }
}
proc update_pouch {} { 
    if {![winfo exists .pouch]} {return}
    set cur [expr {[_current] + 5}]
    if {$cur >= 6} { .pouch.$cur conf -relief raised }
    incr cur; if {$cur < 72} { .pouch.$cur conf -relief sunken }
    incr cur; if {$cur < 72} { .pouch.$cur conf -relief raised }
}
proc create_images {{col 0}} { global img shape fill
    uplevel "#0" {catch {unset img}}; set img() null
    
    foreach s $shape {
        foreach f $fill c {red darkgreen blue purple orange magenta} {
            if {$col} {
                upvar "#0" data_${s}_b data
                image create bitmap shape$s$f -data $data \
                    -foreground $c;# -background white
            } else {
                upvar "#0" data_${s}_${f} data
                image create bitmap shape$s$f -data $data
            }
            set img([list $s $f]) shape$s$f
        }
    }
    foreach i {null cross} { 
        upvar "#0" data_${i} data
        image create bitmap $i -data $data 
    }
}
create_images 0


frame .board -relief ridge -bd 5 -takefocus 1 -highlightthickness 2 
frame .sep -takefocus 1 
frame .cmds -relief sunken -bd 1
frame .options -relief sunken -bd 1 
pack .board .sep -side left
pack .cmds .options -fill both -side top

label .board.bubble -bd 6 -relief ridge -image null
label .board.score -relief sunken -bd 1
label .board.next -bd 1 -relief flat -image null 
label .board.infos -bd 1 -relief sunken 
grid .board.score -column 1 -row 9 -columnspan 12 -sticky nsew
grid .board.next -column 13 -row 9 
grid .board.infos -column 13 -row 1 -rowspan 8 -sticky nsew

button .cmds.quit -text "Quit game" -underline 0 -command {exit}
button .cmds.new -text "new Game" -underline 4 -command {win_new_game}
button .cmds.undo -text "Undo" -underline 0 -command win_undo
button .cmds.redo -text "Redo" -underline 0 -command win_redo
button .cmds.help -text "Help" -underline 0 -command display_help
button .cmds.game -text "game ID" -underline 5 -command display_random
pack .cmds.quit .cmds.new .cmds.undo .cmds.redo .cmds.help .cmds.game \
    -expand no -fill x

checkbutton .options.sound -text "Sound" -variable option(sound)
checkbutton .options.bubble -text "Bubble" \
    -variable option(bubble) -command do_img 
checkbutton .options.pouch -text "Pouch" \
    -variable option(pouch) -command do_pouch 
checkbutton .options.fuzzy -text "Color" \
    -variable option(images) -command {create_images $option(images)}
pack .options.sound .options.bubble .options.pouch \
    .options.fuzzy   -expand no -fill none

foreach x $X {
    foreach y $Y {
        if {$x==1 || $x==12 || $y==1 || $y==8} {
            set color darkgrey 
        } else { set color lightgrey }
            set frame [xy_to_wn $x $y {}]
        set label [xy_to_wn $x $y]
        # frame $frame -highlightthickness 1 -bd 1 -relief sunken
        frame $frame -bd 2 -relief sunken  \
            -highlightthickness 4 -highlightbackground grey 
        label $label -bd 0 -relief sunken -highlightthickness 0
        bindtags $label [concat target [bindtags $label]]
        $label conf -background $color -takefocus 0
        pack $label -expand yes -fill both
        grid $frame -column $x -row $y
        set_image $x $y
    }
}

bind target <1> { win_place %W }
bind all <2> { win_redo }
bind all <3> { win_undo }
bind target <KeyPress> { board_key_handler %W %K }
bind target <Tab> { focus [tk_focusNext .cmds]; break }
bind target <Shift-Tab> { focus [tk_focusPrev .board]; break }
bind all <KeyPress> { key_handler %W %K }
bind .board <FocusIn> { board_focus }
bind .sep <FocusIn> { board_focus }
bind .board.bubble <1> { hide_img }
bind target <FocusIn> { show_img %W }
bind target <FocusOut> { hide_img %W }

set option(sound) 0
set option(bubble) 1
set option(pouch) 0
set option(images) 0

set cur_stone {}
set within_finished 0
set offsx [expr [image width null] - 14]
set offsy [expr [image height null] - 14]

show_stone; board_focus 7 4 
set cur_player "You"
init_player $cur_player; update_score $cur_player

